#include <avr/interrupt.h>
#include <avr/io.h>

#include <util/delay.h>

#include "usart.h"

#define NONE	0
#define LEFT	1
#define RIGHT	2
#define BOTH	3

#define NONE_MOVE	0
#define FRONT_MOVE	1
#define BACK_MOVE	2
#define RIGHT_MOVE	4
#define LEFT_MOVE	8

#define NO_SPEED		0
#define LOW_SPEED		95
#define MEDIUM_SPEED	175
#define HIGH_SPEED		255

volatile static int active_blinker = NONE;
volatile static int active_move = NONE_MOVE;
volatile static int active_speed = NO_SPEED;

static const char LOW_SPEED_COMMAND = 'b';
static const char MEDIUM_SPEED_COMMAND = 'n';
static const char HIGH_SPEED_COMMAND = 'm';
static const char HORN_COMMAND = 'h';
static const char UP_COMMAND = 'w';
static const char DOWN_COMMAND = 's';
static const char RIGHT_COMMAND = 'd';
static const char LEFT_COMMAND = 'a';
static const char STOP_MOVEMENT_COMMAND = 'x';
static const char LIGHTS_COMMAND = 'l';
static const char BLINKER_LEFT_COMMAND = 't';
static const char BLINKER_RIGHT_COMMAND = 'u';
static const char BOTH_BLINKERS_COMMAND = 'y';

/* Functie de setare a vitezei de deplasare */
void set_speed(int level) {
	int last_speed = active_speed;

	if (level == 1) {
		active_speed = LOW_SPEED;	
	} else if (level == 2) {
		active_speed = MEDIUM_SPEED;	
	} else if (level == 3) {
		active_speed = HIGH_SPEED;	
	}

	if (active_move != NONE_MOVE) {
		if (OCR0A == last_speed) {
			OCR0A = active_speed;		
		} else {
			OCR0A = active_speed / 2;		
		}
		if (OCR0B == last_speed) {
			OCR0B = active_speed;		
		} else {
			OCR0B = active_speed / 2;		
		}
	}
}

/* Functia de schimbare a starii (pornit / oprit) claxonului */
void change_horn_state(void) {
	if (OCR2B == 255) {
		// Se porneste claxonul
		OCR2B /= 2;	
	} else {
		// Se opreste claxonul
		OCR2B = 255;	
	}
}

/* Miscare in fata */
void move_front(void) {
	if (active_move == NONE_MOVE && active_speed != NO_SPEED) {
		active_move = FRONT_MOVE;
		PORTB |= (1 << PB0) | (1 << PB1);
		OCR0A = OCR0B = active_speed;
	}
}

/* Miscare inapoi */
void move_back(void) {
	if (active_move == NONE_MOVE && active_speed != NO_SPEED) {
		active_move = BACK_MOVE;
		PORTB &= ~((1 << PB0) | (1 << PB1));
		OCR0A = OCR0B = active_speed;
	}
}

/* Miscare la stanga */
void move_left(void) {
	if (active_move & LEFT_MOVE) {
		active_move -= LEFT_MOVE;
		OCR0B = active_speed;	
	} else {
		if (active_move == FRONT_MOVE || active_move == BACK_MOVE) {
			active_move |= LEFT_MOVE;
			OCR0B = active_speed / 2;	
		}
	}
}


/* Miscare la dreapta */
void move_right(void) {
	if (active_move & RIGHT_MOVE) {
		active_move -= RIGHT_MOVE;
		OCR0A = active_speed;	
	} else {
		if (active_move == FRONT_MOVE || active_move == BACK_MOVE) {
			active_move |= RIGHT_MOVE;
			OCR0A = active_speed / 2;	
		}
	}
}

/* Masina se opreste */
void stop_movement(void) {
	active_move = NONE_MOVE;
	active_speed = NO_SPEED;
	OCR0A = OCR0B = active_speed;
}

/* Functia de schimbare a starii (aprinse / stinse) luminilor */
void change_lights_state(void) {
	PORTA ^= (1 << PA1); 
}

/* Intrerupere pentru semnalizarea stanga */
ISR(TIMER1_COMPA_vect) {
	if (active_blinker == LEFT || active_blinker == BOTH) {
		PORTA ^= (1 << PA0);	
	}
}

/* Intrerupere pentru semnalizare dreapta */
ISR(TIMER1_COMPB_vect) {
	if (active_blinker == RIGHT || active_blinker == BOTH) {
		PORTA ^= (1 << PA2);	
	}
}

/* Functie de schimbare a starii semnalizarii */
void change_blinker_state(int blinker) {
	if (active_blinker == NONE) {
		active_blinker = blinker;
		if (blinker == LEFT) {
			OCR1A = ICR1 / 2;		
		} else if (blinker == RIGHT) {
			OCR1B = ICR1 / 2;		
		}
	} else if (active_blinker == LEFT) {
		if (blinker == LEFT) {
			PORTA &= ~(1 << PA0);
			active_blinker = NONE;
			OCR1A = 0;		
		}	
	} else if (active_blinker == RIGHT) {
		if (blinker == RIGHT) {
			PORTA &= ~(1 << PA2);
			active_blinker = NONE;
			OCR1B = 0;		
		}	
	}
}

/* Functie de schimbarea a starii avariilor */
void change_both_blinkers_state(void) {
	if (active_blinker == BOTH) {
		active_blinker = NONE;
		PORTA &= ~((1 << PA0) | (1 << PA2));
		OCR1A = OCR1B = 0;
	} else {
		active_blinker = BOTH;
		PORTA &= ~((1 << PA0) | (1 << PA2));
		OCR1A = OCR1B = ICR1 / 2;
	}
}

/* Functia de initializare a luminilor */
void init_lights(void) {
	DDRA |= (1 << PA1);
}

/* Functie de initializare a semnalizarilor */
void init_blinkers(void) {
	DDRA |= (1 << PA0) | (1 << PA2);

	ICR1 = 31500;

	/* Semnalizarile sunt stinse initial */
	OCR1A = OCR1B = 0;

	/* CTC cu top la ICR, prescaler de 256 */
	TCCR1B |= (3 << WGM12) | (4 << CS10);

	/* Activez intreruperile */
	TIMSK1 |= (1 << OCIE1A) | (1 << OCIE1B);
}

/* Functia de initializare a buzzer-ului pentru comandarea acestuia prin PWM */
void init_buzzer(void) {
	DDRD |= (1 << PD6);

	// Buzzer-ul este stins initial
	OCR2B = 255;

	// Fast PWM cu top la 0xFF si modul non-inverting
	TCCR2A |= (1 << COM2B1) | (3 << WGM20); 
	// Prescaler la 128
	TCCR2B |= (5 << CS20);	
}

/* Functie de initializare a motoarelor */
void init_motors(void) {
	/* Phase DRV8835 */
	DDRB |= (1 << PB0) | (1 << PB1);

	/* Enable DRV8835 */
	DDRB |= (1 << PB3) | (1 << PB4);

	OCR0A = 0;
	OCR0B = 0;
	
	TCCR0A |= (1 << COM0A1) | (1 << COM0B1) | (3 << WGM00);
	TCCR0B |= (1 << CS00);
}

int main(void)
{
	char received_char;

	sei();

	DDRD |= (1 << PD7);

	USART0_init();
	init_lights();
	init_blinkers();
	init_buzzer();
	init_motors();

	while(1) {
		received_char = USART0_receive();
		switch (received_char)
		{
			case LOW_SPEED_COMMAND:
				set_speed(1);
				break;
			case MEDIUM_SPEED_COMMAND:
				set_speed(2);
				break;
			case HIGH_SPEED_COMMAND:
				set_speed(3);
				break;
			case HORN_COMMAND:
				change_horn_state();
				break;
			case UP_COMMAND:
				move_front();
				break;
			case DOWN_COMMAND:
				move_back();
				break;
			case RIGHT_COMMAND:
				move_right();
				break;
			case LEFT_COMMAND:
				move_left();
				break;
			case STOP_MOVEMENT_COMMAND:
				stop_movement();
				break;
			case LIGHTS_COMMAND:
				change_lights_state();
				break;
			case BLINKER_LEFT_COMMAND:
				change_blinker_state(LEFT);
				break;
			case BLINKER_RIGHT_COMMAND:
				change_blinker_state(RIGHT);
				break;
			case BOTH_BLINKERS_COMMAND:
				change_both_blinkers_state();				
				break;
			default:
				PORTD ^= (1 << PD7);
				_delay_ms(500);
		}
	}

    return 0;
}
